@@ -1,6 +1,6 @@
 
     
-############        CCTfrom***()     ###################    
+############        CCTfrom***()     ###################
     
 #   XYZ     an Nx3 matrix, or a vector that can be converted to such a matrix
 #   returns CCT, or NA if outside valid range
@@ -38,21 +38,48 @@ CCTfromuv  <- function( uv, isotherms='robertson', locus='robertson', strict=FAL
     uv  = prepareNxM( uv, M=2 )
     if( is.null(uv) )   return(NULL)
 
-    isofull = c( 'native.Rob', 'Robertson', 'McCamy' )
+    #   process isotherms    
+    if( length(isotherms) == 0 )
+        {
+        log.string( ERROR, "isotherms is invalid, because length(isotherms)=0."  )
+        return( NULL )
+        }      
+        
+    isofull = c( 'native', 'Robertson', 'McCamy' )
+        
+    isotherms   =  as.character(isotherms)
+    
+    isotherms[ is.na(isotherms) ]   = 'native'
         
     idx.isotherms   = pmatch( tolower(isotherms), tolower(isofull), nomatch=0, duplicates.ok=TRUE )
-    if( length(idx.isotherms)==0  ||  any( idx.isotherms==0 ) )
+    if( any( idx.isotherms==0 ) )
         {
-        log.string( ERROR, "isotherms='%s' is invalid.", as.character(isotherms) )
+        log.string( ERROR, "isotherms='%s' is invalid.", isotherms[idx.isotherms==0] )
         return( NULL )
         }            
         
-    idx.locus   = pmatch( tolower(locus), c('robertson'), nomatch=0 )
+    #   process locus
+    ok  = is.character(locus)  &&  length(locus)==1
+    if( ! ok )
+        {
+        log.string( ERROR, "Argument locus is invalid.  It must be a character vector with length 1." )
+        return(NULL)
+        }
+
+    locusfull = c( 'Robertson', 'precision' )        
+                     
+    idx.locus   = pmatch( tolower(locus), tolower(locusfull), nomatch=0 )
     if( idx.locus == 0 )
         {
         log.string( ERROR, "locus='%s' is invalid.", as.character(locus) )
         return( NULL )
-        }               
+        }            
+
+    if( idx.locus == 1 )
+        locus.list  = p.uvCubicsfromMired
+    else
+        locus.list  = p.uvQuinticsfromMired
+        
 
     n   = nrow(uv)
     out = matrix( NA_real_, n, length(idx.isotherms) )
@@ -69,13 +96,13 @@ CCTfromuv  <- function( uv, isotherms='robertson', locus='robertson', strict=FAL
             {
             #   native Robertson spline
             for( i in 1:n )
-                out[i,j]  = CCTfromuv_native( uv[i, ], locus=locus, strict=strict  )
+                out[i,j]  = CCTfromuv_native( uv[i, ], locus.list, strict  )
             }
         else if( idx == 2 )
             {
             #   Robertson isotherms
             for( i in 1:n )
-                out[i,j]  = CCTfromuv_Robertson( uv[i, ], locus=locus, strict=strict  )
+                out[i,j]  = CCTfromuv_Robertson( uv[i, ], locus.list, strict  )
             }        
         else if( idx == 3 )
             {
@@ -88,7 +115,7 @@ CCTfromuv  <- function( uv, isotherms='robertson', locus='robertson', strict=FAL
                 if( is.na(denom) ||  denom < 1.e-16 )   next
 
                 xy      = c( 1.5*uv[i,1], uv[i,2] ) / denom            
-                out[i,j]  = CCTfromxy_McCamy( xy, locus=locus, strict=strict )
+                out[i,j]  = CCTfromxy_McCamy( xy, locus.list, strict )
                 }
             }
         }
@@ -105,20 +132,24 @@ CCTfromuv  <- function( uv, isotherms='robertson', locus='robertson', strict=FAL
     }
 
     
-#   uv      point, usually off locus
-#   locus   ignored
-#   strict  check distance to locus
+#   uv          point, usually off locus
+#   locus.list  ufun, vfun, and miredInterval
+#   strict      check distance to locus
 #
 #   computes where the native isotherm through uv intersects the locus,
 #   and returns the native temperature there.
 #   It works by searching along the locus, while 'sweeping around' the normal line
 #   to find the normal line passing through uv.
 
-CCTfromuv_native  <- function( uv, locus, strict )
+CCTfromuv_native  <- function( uv, locus.list, strict )
     {
     if( any( is.na(uv) ) )  return(NA_real_)
     
-    if( TRUE )
+    #   log.string( TRACE, "uv=%g,%g", uv[1], uv[2] )
+    #   print( str(locus.list) )
+    
+    
+    if( FALSE )
         {
         #   not sure whether this special case is worth the time
         #   see whether uv is actually in the LUT !   When it happens, the result is impressive and skips all below.
@@ -134,18 +165,18 @@ CCTfromuv_native  <- function( uv, locus, strict )
         {
         #   mir         = 1.e6 / temp
         
-        uv.locus    = c( p.uvfromMired[[1]]( mir ), p.uvfromMired[[2]]( mir ) )
+        uv.locus    = c( locus.list$ufun( mir ), locus.list$vfun( mir ) )
         
-        tangent     = c( p.uvfromMired[[1]]( mir, deriv=1 ), p.uvfromMired[[2]]( mir, deriv=1 ) )
+        tangent     = c( locus.list$ufun( mir, deriv=1 ), locus.list$vfun( mir, deriv=1 ) )
         
         return( sum( tangent * (uv.locus - uv) ) )  # relevant dot product
         }
     
-    intervalMired    = range( p.dataCCT$mired )
+    miredInterval    = locus.list$miredInterval           # range( p.dataCCT$mired )
     
     #   check endpoint values
-    f1  = myfun( intervalMired[1], uv )
-    f2  = myfun( intervalMired[2], uv )
+    f1  = myfun( miredInterval[1], uv )
+    f2  = myfun( miredInterval[2], uv )
     
     if( 0 < f1 * f2 )    
         {
@@ -155,16 +186,16 @@ CCTfromuv_native  <- function( uv, locus, strict )
         }
         
     if( f1 == 0 )
-        mired.end = intervalMired[1]
+        mired.end = miredInterval[1]
     else if( f2 == 0 )
-        mired.end = intervalMired[2]
+        mired.end = miredInterval[2]
     else 
         {
         #log.string( DEBUG, "stats::uniroot() on interval [%g,%g],    endpoint values %g and %g.", 
-        #                        intervalMired[1], intervalMired[2],  f1, f2 )
+        #                        miredInterval[1], miredInterval[2],  f1, f2 )
         
         #   reduced the tolerance here, but still only takes about 8 iterations
-        res = try( stats::uniroot( myfun, interval=intervalMired, uv=uv, tol=.Machine$double.eps^0.5 ),  silent=FALSE )
+        res = try( stats::uniroot( myfun, interval=miredInterval, uv=uv, tol=.Machine$double.eps^0.5 ),  silent=FALSE )
         
         if( class(res) == "try-error" )    
             {
@@ -182,7 +213,7 @@ CCTfromuv_native  <- function( uv, locus, strict )
     if( strict )
         {
         #   mired.end   = 1.e6 / temp.end
-        uv.locus    = c( p.uvfromMired[[1]]( mired.end ), p.uvfromMired[[2]]( mired.end ) )        
+        uv.locus    = c( locus.list$ufun( mired.end ),locus.list$vfun( mired.end ) )        
         resid       = uv - uv.locus
         dist        = sqrt( sum(resid^2) )
         
@@ -206,61 +237,19 @@ CCTfromuv_native  <- function( uv, locus, strict )
 #
 #   requires private data frame p.dataCCT,  which is lazy-loaded from sysdata.rda;  see savePrivateDatasets()
     
-CCTfromuv_Robertson  <- function( uv, locus, strict )
+CCTfromuv_Robertson  <- function( uv, locus.list, strict )
     {
     if( any( is.na(uv) ) )  return(NA_real_)
     
-    idx.locus  = pmatch( tolower(locus), c('robertson'), nomatch=0 )
-    if( idx.locus == 0 )
-        {
-        log.string( ERROR, "locus='%s' is invalid.", as.character(locus) )
-        return(  NA_real_ )
-        }
-
-    di = (uv[2] - p.dataCCT$v) - p.dataCCT$t * (uv[1] - p.dataCCT$u)
-    #   print( di )
+    mired   = miredfromuv_Robertson_nocheck( uv, extrap=FALSE )
     
-    n = length(di)
-    
-    #   di should be decreasing, and with exactly 1 zero crossing
-    i   = 0
-    
-    #   check for an exact 0
-    idx = which( di == 0 )
-    
-    if( 0 < length(idx) )
-        {
-        #   exactly 1 is OK
-        if( length(idx) == 1 )  i = max( idx[1], 2 )
-        }
-    else
-        {
-        #   check for a true crossing
-        idx = which( di[1:(n-1)] * di[2:n] < 0 ) + 1
-        
-        #   exactly 1 is OK
-        if( length(idx) == 1 )  i = idx[1]      #     ||  i==1 ) ?
-        }
-        
-    if( i == 0 )
-        {
-        log.string( WARN, "uv=%g,%g is in invalid region. Cannot find unique zero crossing. CCT cannot be calculated.", uv[1], uv[2] )
-        return( NA_real_ )
-        }
-        
-    d0  = di[i]   / sqrt( 1 + p.dataCCT$t[i]^2 )
-    dm  = di[i-1] / sqrt( 1 + p.dataCCT$t[i-1]^2 )
-    p   = dm / (dm - d0)
-    
-    mired   = (1-p)*p.dataCCT$mired[i-1]  +  p*p.dataCCT$mired[i]
-    
-    #   offset   = (1-p) * c(p.dataCCT$u[i-1],p.dataCCT$v[i-1])  +  p * c(p.dataCCT$u[i],p.dataCCT$v[i])  - uv
+    if( is.na(mired) )  return(NA_real_)
     
     CCT = 1.e6 / mired                
     
     if( strict )
         {
-        res = nativeFromRobertson( CCT, locus )
+        res = nativeFromRobertson( CCT, locus.list )
         
         if( is.null(res) )  return( NA_real_ )
         
@@ -279,8 +268,75 @@ CCTfromuv_Robertson  <- function( uv, locus, strict )
     }
     
     
-#   CCT     temperature defining a Robertson isotherm, in K.   CCT=Inf is OK, mired is then 0.
-#   locus   currently ignored. Using the Robertson spline
+  
+#   uv          a 2-vector with uv 1960 
+#   extrap      TRUE means extrapolate past mired=0 and 600.  FALSE means return NA_real
+#   requires private data frame p.dataCCT,  which is lazy-loaded from sysdata.rda;  see savePrivateDatasets()
+#   'nocheck' means do not that input is NA and do not check locus
+
+miredfromuv_Robertson_nocheck  <- function( uv, extrap=FALSE )
+    {
+    di = (uv[2] - p.dataCCT$v) - p.dataCCT$t * (uv[1] - p.dataCCT$u)
+    #   print( di )
+    
+    n = length(di)
+    
+    #   di should be decreasing, and with exactly 1 zero crossing
+    #   i   = 0
+    
+    #   check for an exact 0
+    idx = which( di == 0 )
+    
+    if( 0 < length(idx) )
+        {
+        if( length(idx) == 1 )  
+            #   exactly 1 is OK        
+            return( p.dataCCT$mired[idx] )
+        
+        #   more than 1 should not happen
+        log.string( WARN, "uv=%g,%g lies on more than 1 isotherm. Mired cannot be calculated.", uv[1], uv[2] )
+        return( NA_real_ )
+        }
+
+    #   check for a true crossing
+    idx = which( di[1:(n-1)] * di[2:n] < 0 )
+    
+    if( length(idx) == 0 )    
+        {
+        #   all di must be pos or neg
+        if( ! extrap )
+            {
+            log.string( WARN, "uv=%g,%g is in invalid region. Found no zero-crossings. Mired cannot be calculated.", uv[1], uv[2] )
+            return( NA_real_ )
+            }        
+            
+        #   do a very crude extrapolation, only useful for root-finding
+        if( di[1] < 0 )
+            return( p.dataCCT$mired[1] + 100 * di[1] )
+        else
+            return( p.dataCCT$mired[n] + 30 * di[n] )
+        }
+    else if( 2 <= length(idx) )        #     ||  i==1 ) ?
+        {
+        log.string( WARN, "uv=%g,%g is in invalid region. Found multiple zero-crossings. Mired cannot be calculated.", uv[1], uv[2] )
+        return( NA_real_ )
+        }        
+        
+    #   exactly 1 zero-crossing is OK.  This is the normal situation        
+    i = idx[1] + 1
+    
+    d0  = di[i]   / sqrt( 1 + p.dataCCT$t[i]^2 )
+    dm  = di[i-1] / sqrt( 1 + p.dataCCT$t[i-1]^2 )
+    p   = dm / (dm - d0)
+    
+    mired   = (1-p)*p.dataCCT$mired[i-1]  +  p*p.dataCCT$mired[i]
+        
+    return( mired )
+    }
+    
+    
+#   CCT         temperature defining a Robertson isotherm, in K.   CCT=Inf is OK, mired is then 0.
+#   locus.list  ufun, vfun, and miredInterval
 #
 #   find where the isotherm intersects the locus, using uniroot()
 #   in general, we are looking at the spline locus between isotherm i and isotherm i+1; a sector.
@@ -289,22 +345,23 @@ CCTfromuv_Robertson  <- function( uv, locus, strict )
 #           uv      the point of intersection on the locus
 #           CCT     the native parameterization CCT in K
 #           normal  unit vector along the isotherm, and approximate normal to locus
+#                   this is used to translate the point uv a distance delta away from the locus
 #   Or NULL in case of error.
 
-nativeFromRobertson <- function( CCT, locus )
+nativeFromRobertson <- function( CCT, locus.list )
     {    
     mired   = 1.e6 / CCT
     
-    i   = findInterval( mired, p.dataCCT$mired )
+    j   = findInterval( mired, p.dataCCT$mired )
     
-    if( i == 0 )
+    if( j == 0 )
         {
         log.string( WARN, "CCT=%g is out of the Robertson LUT range. mired < %g", CCT, p.dataCCT$mired[1] )
         return(NULL)
         }
             
     n   = length( p.dataCCT$mired )
-    if( i == n )
+    if( j == n )
         {
         #   an extra check. 
         #   although I have found that 1.e6 / (1.e6 / 600) == 600 exactly on my PC, this might not be TRUE on other hardware,
@@ -312,101 +369,112 @@ nativeFromRobertson <- function( CCT, locus )
         epsilon = 1.e-12
         if( p.dataCCT$mired[n] + epsilon  <  mired )
             {
-            log.string( WARN, "CCT=%g is outside the Robertson LUT range. %g < mired", CCT, p.dataCCT$mired[n] )
+            log.string( WARN, "CCT=%g is outside the Robertson LUT range. %g < %g mired", 
+                            CCT, p.dataCCT$mired[n], mired )
             return(NULL)
             }     
 
         #   otherwise set mired to what it probably should be (600),
-        #   and keeping going with i == n, and the next special if block will take care of it !
+        #   and keeping going with j == n, and the next special if block will take care of it !
         mired = p.dataCCT$mired[n] 
         }
 
     out = list()
     
-    if( p.dataCCT$mired[i] == mired )    
+    if( p.dataCCT$mired[j] == mired )    
         {
-        #   special case, CCT is at one of the defining isotherms.  root at endpoint so avoid using uniroot()
-        #log.string( DEBUG, "mired=%g exactly. isotherm %d.  uniroot unnecessary.", mired, i )
+        #   special case, CCT is at one of the defining isotherms.
+        #   The root is at the endpoint, so avoid using uniroot().
+        #log.string( DEBUG, "mired=%g exactly. isotherm %d.  uniroot unnecessary.", mired, j )
         
-        out$uv  = c( p.dataCCT$u[i], p.dataCCT$v[i] )
+        out$uv  = c( p.dataCCT$u[j], p.dataCCT$v[j] )
         out$CCT = CCT      
         
-        tani        = c( 1, p.dataCCT$t[i] )    # tangent vector to isotherm
-        len         = sqrt( sum(tani^2) )
-        out$normal  = -tani / len               # normal to locus (approx)
+        tanj        = c( 1, p.dataCCT$t[j] )    # tangent vector to isotherm
+        len         = sqrt( sum(tanj^2) )
+        out$normal  = -tanj / len               # normal to locus (approx)
         
         return(out)
         }
         
-    #   we are looking at the spline locus between isotherm i and isotherm i+1. a sector.
+    #   we are looking at the spline locus between isotherm j and isotherm j+1. a sector.
     #   See Fig. 2(3.11) in W&S. 
+    
+    #   compute fraction of the way between j and j+1
+    alpha   =   (mired - p.dataCCT$mired[j]) / (p.dataCCT$mired[j+1] - p.dataCCT$mired[j])
         
-    #   extract the 2 points
-    uvi     = c( p.dataCCT$u[i], p.dataCCT$v[i] )
-    uvi1    = c( p.dataCCT$u[i+1], p.dataCCT$v[i+1] )
+    #   extract the 2 points on isotherms j and j+1
+    pj      = c( p.dataCCT$u[j], p.dataCCT$v[j] )
+    pj1     = c( p.dataCCT$u[j+1], p.dataCCT$v[j+1] )
       
-    #   precompute unit normals to the isotherms at both ends.  Maybe these should be moved into the LUT someday.
+    #   compute unit normals to the isotherms at both ends.  Maybe these should be moved into the LUT someday.
     #   these are very close to the unit tangents to the locus.
-    tani    = c( 1, p.dataCCT$t[i] )            # tangent vector to isotherm
-    len     = sqrt( sum(tani^2) )
-    normi   = c( -tani[2], tani[1] ) / len      # unit normal to isotherm at knot i
+    tanj    = c( 1, p.dataCCT$t[j] )            # tangent vector to isotherm
+    len     = sqrt( sum(tanj^2) )
+    normj   = c( -tanj[2], tanj[1] ) / len      # unit normal to isotherm j
     
-    tani1   = c( 1, p.dataCCT$t[i+1] )          # tangent vector to isotherm
-    len     = sqrt( sum(tani1^2) )
-    normi1  = c( -tani1[2], tani1[1] ) / len    # unit normal to isotherm at knot i+1
-        
-    myfun   <- function( mired, CCT )
+    tanj1   = c( 1, p.dataCCT$t[j+1] )          # tangent vector to isotherm
+    len     = sqrt( sum(tanj1^2) )
+    normj1  = c( -tanj1[2], tanj1[1] ) / len    # unit normal to isotherm j+1
+
+    #   compute the normal for the intermediate isotherm
+    norm    = (1-alpha)*normj  +  alpha*normj1
+    #   norm    = norm / sqrt( sum(norm^2) )        do not make it a unit normal
+    
+    #   compute point p on the intermediate isotherm, on the segment [pj,pj1]
+    delta   = pj1 - pj
+    s       = alpha * sum( delta * normj1 ) / sum( delta * norm )
+    p       = pj  +  s*delta
+    
+    #   from now on, mired refers to the parameter on locus.list, and not the Robertson mired isotherm.
+    
+    myfun   <- function( mired )
         {
-        uv  = c( p.uvfromMired[[1]]( mired ), p.uvfromMired[[2]]( mired ) )
+        uv  = c( locus.list$ufun( mired ), locus.list$vfun( mired ) )
         
-        deltai  = uv - uvi
-        di      = sum( deltai*normi )
-        deltai1 = uv - uvi1
-        di1     = sum( deltai1*normi1 )
-        
-        denom   = di - di1
-        
-        ok  = (0 <= di)  &&  (di1 <= 0) && (0 < denom)
-        if( ! ok )
-            log.string( ERROR, "Bad signs. di=%g and di1=%g. mired=%g. i=%d", di, di1, mired, i )
-    
-        alpha   = di / denom   # this must be in closed interval [0,1]
-    
-        mir = (1-alpha)*p.dataCCT$mired[i]  +  alpha*p.dataCCT$mired[i+1]
-            
-        return( 1.e6/mir - CCT )
+        return( sum( (uv - p)*norm ) )
         }
     
-    rangeMired    = c( p.dataCCT$mired[i], p.dataCCT$mired[i+1] )
+    miredInterval   = locus.list$miredInterval         #c( p.dataCCT$mired[i], p.dataCCT$mired[i+1] )
 
-    #log.string( DEBUG, "stats::uniroot() on interval [%g,%g], isotherms=%d and %d.  endpoint values %g and %g.", 
-    #                        rangeMired[1], rangeMired[2], i, i+1, myfun(rangeMired[1],CCT), myfun(rangeMired[2],CCT)  )
-
-    res = try( stats::uniroot( myfun, interval=rangeMired, CCT ),  silent=FALSE )
+    #   check endpoint values
+    f1  = myfun( miredInterval[1] )
+    f2  = myfun( miredInterval[2] )
     
-    if( class(res) == "try-error" )    
+    if( 0 < f1 * f2 )    
         {
-        cat( 'stats::uniroot()  res = ', utils::str(res), '\n', file=stderr() )
+        #   same sign, should not happen
+        log.string( WARN, "CCT=%g  mired=%g  p=%g,%g.  norm=%g,%g. Test function has the same sign at endpoints [%g,%g]. %g and %g CCT cannot be calculated.", 
+                                CCT, mired, p[1], p[2], norm[1], norm[2], miredInterval[1], miredInterval[2], f1, f2  )
         return( NULL )
-        }        
+        }
+        
+    if( f1 == 0 )
+        mired.end = miredInterval[1]
+    else if( f2 == 0 )
+        mired.end = miredInterval[2]    
+    else
+        {
+        #log.string( DEBUG, "stats::uniroot() on interval [%g,%g], isotherms=%d and %d.  endpoint values %g and %g.", 
+        #                        rangeMired[1], rangeMired[2], i, i+1, myfun(rangeMired[1],CCT), myfun(rangeMired[2],CCT)  )
+
+        res = try( stats::uniroot( myfun, interval=miredInterval ),  silent=FALSE )
+        
+        if( class(res) == "try-error" )    
+            {
+            cat( 'stats::uniroot()  res = ', utils::str(res), '\n', file=stderr() )
+            return( NULL )
+            }     
+
+        mired.end = res$root
+        }
         
     #log.string( DEBUG, "stats::uniroot() successful after %d iterations.  isotherm %d", res$iter, i )
 
-    out$uv  = c( p.uvfromMired[[1]]( res$root ), p.uvfromMired[[2]]( res$root ) )
-    out$CCT = 1.e6 / res$root
-    
-    #   for the unit normal go through distance calculation and interpolation again !
-    deltai  = out$uv - uvi
-    di      = sum( deltai*normi )
-    deltai1 = out$uv - uvi1
-    di1     = sum( deltai1*normi1 )
-    denom   = di - di1    
-    alpha   = di / denom   # this must be in closed interval [0,1]
-    tangent = (1-alpha)*normi  +  alpha*normi1
-    normal  = c( -tangent[2], tangent[1] )      # tangent to locus
-    len     = sqrt( sum(normal^2) )
-    
-    out$normal  = normal / len      #; print( out$normal )
+    out$uv      = c( locus.list$ufun( mired.end ), locus.list$vfun( mired.end ) )
+    out$CCT     = 1.e6 / mired.end
+    norm        = norm / sqrt( sum(norm^2) )    # make intermediate norm a unit vector     
+    out$normal  = c( -norm[2], norm[1] )        # normal / len      #; print( out$normal )
             
     return( out )
     }
@@ -426,16 +494,16 @@ CCTfromxy_McCamy_nocheck  <- function( xy )
     
     if( topbot[2] <= 0 )  return( NA_real_ )
     
-    n   = topbot[1]/topbot[2]
+    w   = topbot[1]/topbot[2]
     
-    out = -449*n^3 + 3525*n^2 - 6823.3*n + 5520.33
+    out = ((-449*w + 3525)*w - 6823.3)*w  +  5520.33
     
-    if( out <= 0 )  return( NA_real_ )  # this can happen
+    if( out <= 0 )  return( NA_real_ )  # this *can* happen if w is too big
     
     return( out )
     }
     
-CCTfromxy_McCamy  <- function( xy, locus, strict )
+CCTfromxy_McCamy  <- function( xy, locus.list, strict )
     {
     if( any( is.na(xy) ) )  return( NA_real_ )
     
@@ -444,7 +512,7 @@ CCTfromxy_McCamy  <- function( xy, locus, strict )
     if( is.finite(CCT)  &&  strict )
         {
         #   find intersection of McCamy isotherm and the locus
-        res = nativeFromMcCamy( CCT, locus )
+        res = nativeFromMcCamy( CCT, locus.list )
         
         if( is.null(res) )  return( NA_real_ )
         
@@ -464,7 +532,102 @@ CCTfromxy_McCamy  <- function( xy, locus, strict )
     }
     
     
-############        planckLocus()     ###################    
+#   CCT         temperature defining a McCamy isotherm, in K
+#   locus.list  ufun, vfun, and miredInterval
+#
+#   find where the isotherm intersects the locus, using uniroot()
+#
+#   return  a list with:
+#           uv      the point of intersection on the locus
+#           CCT     the native parameterization CCT in K
+#           normal  to the locus, always along the McCamy isotherm for CCT
+#       or NULL in case of error.
+nativeFromMcCamy <- function( CCT, locus.list )
+    {
+    if( 34530 < CCT )   # the McCamy isotherm just below CCT=Inf.  34539
+        return(NULL)
+        
+    if( CCT < 1621 )    # the cubic polynomial 'turns around' at about 1620.245
+        return(NULL)
+        
+    ifun    <- function( w )    { ((-449*w + 3525)*w - 6823.3)*w + 5520.33 - CCT }
+        
+    #   the following interval comes from playing around with the cubic
+    res = try( stats::uniroot( ifun, interval=c(-1.91,1.28), tol=.Machine$double.eps^0.33 ), silent=FALSE )
+    
+    if( class(res) == "try-error" )    
+        {
+        cat( 'stats::uniroot()  res = ', utils::str(res), '\n', file=stderr() )
+        return( NULL )
+        }        
+    # log.string( DEBUG, "stats::uniroot() inverted McCamy cubic successful after %d iterations.", res$iter )
+    
+    #   find equation of the McCamy isotherm for this CCT
+    alpha   = res$root        
+    meet    = c(0.3320,0.1858)          #  the (x,y) 1931 where all the isotherms meet    
+    C       = sum( c(1,-alpha) * meet )
+    normal  = c( 1.5 - C, 4*C - alpha ) # normal to the isotherm
+    uv0     = uvfromxy( meet, 1960 )    # all isotherms pass through this point.  uv0 = (0.2908709, 0.2441738)    
+        
+    myfun <- function( mired )
+        {
+        uv  = c( locus.list$ufun(mired), locus.list$vfun(mired) )
+        
+        return( sum( (uv-uv0)*normal ) )              # or ( sum(uv*normal) - 2*C )
+        
+        #xy  = c( 1.5*uv[1], uv[2] ) / ( uv[1] - 4*uv[2] + 2 )    
+        #return( CCTfromxy_McCamy_nocheck(xy) - CCT )
+        }
+        
+    miredInterval    = locus.list$miredInterval    # range( p.dataCCT$mired )
+        
+    #   log.string( DEBUG, "myfun() at endpoints: %g and %g.", myfun(miredInterval[1],CCT), myfun(miredInterval[2],CCT) )
+    
+    #   check endpoint values
+    f1  = myfun( miredInterval[1] )
+    f2  = myfun( miredInterval[2] )
+    
+    if( 0 < f1 * f2 )    
+        {
+        #   same sign, should not happen
+        log.string( WARN, "CCT=%g.  Test function has the same sign at endpoints [%g,%g]. %g and %g. Intersection of isotherm and locus cannot be calculated.", 
+                                CCT,  miredInterval[1], miredInterval[2], f1, f2  )
+        return( NULL )
+        }
+        
+    if( f1 == 0 )
+        mired.end = miredInterval[1]
+    else if( f2 == 0 )
+        mired.end = miredInterval[2]    
+    else
+        {    
+        res = try( stats::uniroot( myfun, interval=miredInterval, tol=.Machine$double.eps^0.33 ),  silent=FALSE )
+        
+        if( class(res) == "try-error" )    
+            {
+            cat( 'stats::uniroot()  res = ', utils::str(res), '\n', file=stderr() )
+            return( NULL )
+            }        
+        # log.string( DEBUG, "stats::uniroot() found mired.end after %d iterations, for McCamy.", res$iter )
+        
+        mired.end   = res$root
+        }
+        
+    uv  = c( locus.list$ufun( mired.end ), locus.list$vfun( mired.end ) )
+    
+    out = list()
+    out$uv  = uv
+    out$CCT = 1.e6 / mired.end
+    out$normal  = c(-normal[2],normal[1]) / sqrt( sum(normal^2) )          #  approximate normal to the locus
+    
+    return(out)
+    }
+    
+    
+    
+    
+    
+############        planckLocus()     ###############################################################
 
 #   temperature a N-vector of temperatures, in K.  Inf is OK
 #   locus       must match 'robertson'
@@ -476,6 +639,7 @@ CCTfromxy_McCamy  <- function( xy, locus, strict )
 #
 planckLocus  <- function( temperature, locus='robertson', param='robertson', delta=0, space=1960 )
     {
+    #   process temperature
     ok  = is.numeric(temperature)  &&  0<length(temperature)
     if( ! ok )
         {
@@ -483,20 +647,49 @@ planckLocus  <- function( temperature, locus='robertson', param='robertson', del
         return(NULL)
         }
 
-    idx.locus   = pmatch( tolower(locus), c('robertson'), nomatch=0 )
+    #   process locus
+    ok  = is.character(locus)  &&  length(locus)==1
+    if( ! ok )
+        {
+        log.string( ERROR, "Argument locus is invalid.  It must be a character vector with length 1." )
+        return(NULL)
+        }
+
+    locusfull = c( 'Robertson', 'precision' )        
+               
+    idx.locus   = pmatch( tolower(locus), tolower(locusfull), nomatch=0 )
     if( idx.locus == 0 )
         {
         log.string( ERROR, "locus='%s' is invalid.", as.character(locus) )
         return( NULL )
+        }    
+
+    if( idx.locus == 1 )
+        locus.list  = p.uvCubicsfromMired
+    else
+        locus.list  = p.uvQuinticsfromMired
+
+        
+    #   process param
+    if( length(param) != 1 )
+        {
+        log.string( ERROR, "param is invalid, because it has length=%d != 1.", length(param)  )
+        return( NULL )
         }        
         
-    idx.param  = pmatch( tolower(param), c('native','robertson','mccamy'), nomatch=0 )
+    param   = as.character(param)
+    param[ is.na(param) ] = 'native'
+        
+    paramfull = c( 'native', 'Robertson', 'McCamy' )        
+        
+    idx.param  = pmatch( tolower(param), tolower(paramfull), nomatch=0 )
     if( idx.param == 0 )
         {
-        log.string( ERROR, "param='%s' is invalid.", as.character(param) )
+        log.string( ERROR, "param='%s' is invalid.", param  )
         return( NULL )
         }
-        
+    
+    #   process delta
     n   = length(temperature)
     
     ok  = is.numeric(delta)  &&  length(delta) %in% c(1,n)
@@ -507,13 +700,14 @@ planckLocus  <- function( temperature, locus='robertson', param='robertson', del
         }
     if( length(delta) == 1 )    delta = rep( delta[1], n )
 
-    
+    #   process space    
     if( ! match(space,c(1960,1976,1931),nomatch=FALSE) )
         {
         log.string( ERROR, "space='%s' is invalid.",  as.character(space[1]) )
         return(NULL)
         }    
     
+    
     uv  = matrix( NA_real_, n, 2 )
     
     rnames  = names(temperature)
@@ -522,19 +716,21 @@ planckLocus  <- function( temperature, locus='robertson', param='robertson', del
     rownames(uv)    = rnames
     colnames(uv)    = c('u','v')        
         
-    # amazingly, 600 seems to involute perfectly with 1.e6, though not with 1        
-    temperaturerange    = c( 1.e6 / p.dataCCT$mired[ length(p.dataCCT$mired) ], Inf )        # 33334  100000 )
-    
-    ok  =   temperaturerange[1]<=temperature  &  temperature<=temperaturerange[2]
-    temperature[ ! ok ] = NA_real_
-
+    if( FALSE )     # now let temperatures out of range be taken care of by lower functions
+        {
+        # amazingly, 600 seems to involute perfectly with 1.e6
+        temperaturerange    = range( 1.e6 / locus.list$miredInterval )        # 33334  100000 )
+        
+        ok  =   temperaturerange[1]<=temperature  &  temperature<=temperaturerange[2]
+        temperature[ ! ok ] = NA_real_
+        }
     
     if( idx.param == 1 )
         {
-        #   the easy one - plain 'native' spline
+        #   the easy one - plain 'native' 
         mired   = 1.e6 / temperature            # temperature==Inf is OK here, mired is then 0
-        uv[ ,1]   = p.uvfromMired[[1]]( mired )
-        uv[ ,2]   = p.uvfromMired[[2]]( mired )
+        uv[ ,1]   = locus.list$ufun( mired )
+        uv[ ,2]   = locus.list$vfun( mired )
         
         if( any( delta!=0 ) )
             {
@@ -544,7 +740,7 @@ planckLocus  <- function( temperature, locus='robertson', param='robertson', del
                 if( delta[i] == 0  ||  is.na(mired[i]) ) next
                 
                 #   compute unit normal to the locus at point i
-                normal  = c( -p.uvfromMired[[2]]( mired[i], deriv=1 ), p.uvfromMired[[1]]( mired[i], deriv=1 ) )
+                normal  = c( -locus.list$vfun( mired[i], deriv=1 ), locus.list$ufun( mired[i], deriv=1 ) )
                 len     = sqrt( sum(normal^2) )
                 if( len == 0 )  next    # should never happen
                 normal  = normal / len
@@ -562,19 +758,19 @@ planckLocus  <- function( temperature, locus='robertson', param='robertson', del
             #if( is.infinite(temperature[i]) )
             #    {
             #    #   special case
-            #    uv[ ,1]   = p.uvfromMired[[1]]( 0 )
-            #    uv[ ,2]   = p.uvfromMired[[2]]( 0 )
+            #    uv[ ,1]   = p.uvCubicsfromMired[[1]]( 0 )
+            #    uv[ ,2]   = p.uvCubicsfromMired[[2]]( 0 )
             #    next
             #    }
             
             if( idx.param == 3 )
                 {
-                res = nativeFromMcCamy( temperature[i], locus )
+                res = nativeFromMcCamy( temperature[i], locus.list )
                 }
             else
                 {
                 #   has to be Robertson
-                res = nativeFromRobertson( temperature[i], locus )     # also works if temperature[i] is Inf
+                res = nativeFromRobertson( temperature[i], locus.list )     # also works if temperature[i] is Inf
                 }
                 
             if( is.null(res) ) next
@@ -589,7 +785,7 @@ planckLocus  <- function( temperature, locus='robertson', param='robertson', del
             }
         }
         
-    #   now handle the space
+    #   now handle the output space
     if( space == 1931 )   
         {
         xy  = uv    # get the size and rownames right
@@ -611,66 +807,3 @@ planckLocus  <- function( temperature, locus='robertson', param='robertson', del
     return(uv)
     }
     
-    
-#   CCT     temperature defining a McCamy isotherm, in K
-#   locus   currently ignored
-#
-#   find where the isotherm intersects the (Robertson) locus, using uniroot()
-#
-#   return  a list with:
-#           uv      the point of intersection on the locus
-#           CCT     the native parameterization CCT in K
-#           normal  to the locus, always along the McCamy isotherm for CCT
-#       or NULL in case of error.
-nativeFromMcCamy <- function( CCT, locus )
-    {
-    #   the following limits were computed from the endpoints of the Robertson lookup table
-    #   and rounded to the appropriate integer
-    ok  = 1702 <= CCT  &&  CCT <= 34539
-    
-    if( ! ok )  return(NULL)    # too small or too big for McCamy           33333
-       
-    myfun <- function( mired, CCT )
-        {
-        uv  = c( p.uvfromMired[[1]](mired), p.uvfromMired[[2]](mired) )
-        
-        xy  = c( 1.5*uv[1], uv[2] ) / ( uv[1] - 4*uv[2] + 2 )    
-        
-        return( CCTfromxy_McCamy_nocheck(xy) - CCT )
-        }
-        
-    rangeMired    = range( p.dataCCT$mired )
-        
-    #   log.string( DEBUG, "myfun() at endpoints: %g and %g.", myfun(rangeMired[1],CCT), myfun(rangeMired[2],CCT) )
-    
-    res = try( stats::uniroot( myfun, interval=rangeMired, CCT ),  silent=FALSE )
-    
-    if( class(res) == "try-error" )    
-        {
-        cat( 'stats::uniroot()  res = ', utils::str(res), '\n', file=stderr() )
-        return( NULL )
-        }        
-    # log.string( DEBUG, "uniroot() successful after %d iterations, for McCamy.", res$iter )
-    
-    uv  = c( p.uvfromMired[[1]]( res$root ), p.uvfromMired[[2]]( res$root ) )
-    
-    out = list()
-    out$uv  = uv
-    out$CCT = 1.e6 /  res$root
-    
-    # normal of locus, along the isotherm
-    xy  = c( 1.5*uv[1], uv[2] ) / ( uv[1] - 4*uv[2] + 2 )    
-        
-    meet    = c(0.3320,0.1858) #  the point where all the isotherms meet
-    topbot  = xy - meet
-    
-    alpha       = topbot[1] / topbot[2]
-    C           = sum( c(1,-alpha) * meet )
-    normal      = c( alpha - 4*C, 1.5 - C )
-    len         = sqrt( sum(normal^2) )
-    out$normal  = normal / len          #; print( out$normal )
-    
-    return(out)
-    }
-    
-    
